#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "cpu.h"
#include "emu.h"
#include "mhpdbg.h"

extern void handle_signals(void);
extern struct vm86_struct vm86s;

static void mhp_poll(void);
static void mhp_send(void);
static void mhp_puts(char*);
static void mhp_putc(char);

static char mhp_banner[] = {"DOSEMU Debugger V0.1 connected\n"};
static struct mhpdbg mhpdbg;
       struct mhpdbgc mhpdbgc;

/********/
/* CODE */
/********/

static void mhp_puts(char* s)

{
   for (;;){
	   if (*s == 0x00)
		   break;
	   mhp_putc (*s++);
   }
}

static void mhp_putc(char c1)

{

   if (c1 == '\n') {
      mhpdbg.sendbuf[mhpdbg.sendptr] = '\r';
      if (mhpdbg.sendptr < SRSIZE-1)
	 mhpdbg.sendptr++;
   }

   mhpdbg.sendbuf[mhpdbg.sendptr] = c1;
   if (mhpdbg.sendptr < SRSIZE-1)
      mhpdbg.sendptr++;

}

static void mhp_send(void)
{
   if ((mhpdbg.sendptr) && (mhpdbg.usocketd == -1)) {
      mhpdbg.sendptr = 0;
      return;
   }
   if ((mhpdbg.sendptr) && (mhpdbg.usocketd != -1)) {
      write (mhpdbg.usocketd, mhpdbg.sendbuf, mhpdbg.sendptr);
      mhpdbg.sendptr = 0;
   }
}

static void mhp_init(void)

{
  mhpdbg.socketd = mhpdbg.usocketd = -1;

  if ((mhpdbg.socketd = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) == -1) {
      return;
  }

  memset ((char *) &mhpdbg.sa_server, 0x00, sizeof(mhpdbg.sa_server));
  mhpdbg.sa_server.sin_family=AF_INET;
  mhpdbg.sa_server.sin_addr.s_addr=htonl(INADDR_ANY);
  mhpdbg.sa_server.sin_port=htons(MYPORT);
  if(bind(mhpdbg.socketd,(struct sockaddr *) &mhpdbg.sa_server,sizeof(mhpdbg.sa_server)) == -1) {
      close(mhpdbg.socketd);
      mhpdbg.socketd = -1;
      return;
  }
  if (listen(mhpdbg.socketd,5) == -1) {
      close(mhpdbg.socketd);
      mhpdbg.socketd = -1;
      return;
  }
  if (ioctl (mhpdbg.socketd, FIONBIO, 1) == -1) {
      close(mhpdbg.socketd);
      mhpdbg.socketd = -1;
      return;
  }

}

static void mhp_poll(void)
{
   int clen;
   int nbytes;
   int rc;
   int i;

   if (mhpdbg.socketd == -1)
      return;

   clen = sizeof(mhpdbg.sa_client);
   if (mhpdbg.usocketd == -1) {
      if((mhpdbg.usocketd=accept(mhpdbg.socketd,&mhpdbg.sa_client,&clen)) == -1)
	  return;
      /* new session has started */
      if (ioctl (mhpdbg.usocketd, FIONBIO, 1) == -1) {
	  close(mhpdbg.usocketd);
	  mhpdbg.usocketd = -1;
	  return;
      }
      mhp_printf ("%s", mhp_banner);
      mhp_cmd("rmapfile");
      mhp_cmd("r");
      mhp_send();
   }

   if (mhpdbgc.stopped) {
      mhp_cmd("r");
      mhp_send();
   }

   for (;;) {
      handle_signals();
      if(ioctl(mhpdbg.usocketd, FIONREAD, &nbytes) == -1) {
	  close(mhpdbg.usocketd);
	  mhpdbg.usocketd = -1;
	  return;
      }
      if (nbytes == 0) {
	 if (!mhpdbgc.stopped)
	    break;
	 else
	    continue;
      }
      if ((nbytes = read(mhpdbg.usocketd, mhpdbg.recvbuf, SRSIZE)) == -1) {
	  close(mhpdbg.usocketd);
	  mhpdbg.usocketd = -1;
	  return;
      }
      if (nbytes > 0) {
	 if ((mhpdbg.recvbuf[0] == 'q') && (mhpdbg.recvbuf[1] <= ' ')) {
	    close(mhpdbg.usocketd);
	    mhpdbg.usocketd = -1;
	    return;
	 }
	 mhpdbg.recvbuf[nbytes] = 0x00;
	 mhp_cmd(mhpdbg.recvbuf);
	 mhp_send();
      }
   }
}

unsigned int mhp_debug(unsigned int code, unsigned int parm1, unsigned int parm2)
{
  int rtncd = 0;
#if 0
  return rtncd;
#endif
  mhpdbgc.currcode = code;
  mhp_bpclr();
  switch (DBG_TYPE(mhpdbgc.currcode)) {
  case DBG_INIT:
	  mhp_init();
	  break;
  case DBG_INTx:
	  if (mhpdbg.usocketd == -1)
	     break;
	  if (mhpdbgc.intxxtab[DBG_ARG(mhpdbgc.currcode)]) {
		  mhpdbgc.stopped = 1;
		  mhp_poll();
	  }
	  break;
  case DBG_TRAP:
	  if (mhpdbg.usocketd == -1)
	     break;
	  if (DBG_ARG(mhpdbgc.currcode) == 1) { /* single step */
		  if (mhpdbgc.trapcmd) {
			  mhpdbgc.trapcmd = 0;
			  rtncd = 1;
			  mhpdbgc.stopped = 1;
			  mhp_poll();
		  }
	  }
	  if (DBG_ARG(mhpdbgc.currcode) == 3) { /* int3 (0xCC) */
		  if (mhp_bpchk( (unsigned char *)((LWORD(cs) << 4) + LWORD(eip) - 1))) {
			  LWORD(eip) --;
			  mhpdbgc.trapcmd = 0;
			  rtncd = 1;
			  mhpdbgc.stopped = 1;
			  mhp_poll();
		  }
	  }
	  break;
  case DBG_POLL:
	  mhp_poll();
	  break;
  case DBG_GPF:
	  if (mhpdbg.usocketd == -1)
	     break;
	  mhpdbgc.stopped = 1;
	  mhp_poll();
	  break;
  default:
	  break;
  }
  if (mhpdbg.usocketd != -1)
     mhp_bpset();
  return rtncd;
}

void mhp_printf(const char *fmt,...)
{
  va_list args;
  char frmtbuf[SRSIZE];

  va_start(args, fmt);
  vsprintf(frmtbuf, fmt, args);
  va_end(args);

  mhp_puts(frmtbuf);
}
